/**
 *
 * Copyright (C) 2013 Olivier Berthier
 *                    http://www.xunil.fr
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 **/
#include <cmath>   
#include <stdlib.h> 
#include <stdio.h>  
#include <vector>
#include <ctime>
#include <pthread.h>
#include <unistd.h>



#include "Ray.h"
#include "Node.h"

#include "And.h"
#include "Or.h"
#include "Sub.h"

#include "Sphere.h"
#include "Box.h"
#include "Cone.h"
#include "Cylinder.h"
#include "Tore.h"

#include "Parameters.h"
#include "Intersections.h"

#include "Matrice.h"

#include "json.h"
    

#define NB_THREAD 8

#define X (240/2) // Width
#define Y (240/2) // Height
#define S 10     // Samples
#define D 1       // Recursion depth


typedef struct
{
  Vector *c;
  pthread_t thread_render;
  pthread_t thread_renders [NB_THREAD];
 
  //   pthread_mutex_t mutex_stock;
}
  render_t;
 
typedef struct str_thdata
{
  int thread_no;
} thdata;

static render_t render; // TODO : variable global => bof

//! \brief Easy coded random function
inline double Random() { return (rand()%10000)/9999.0;}

//!< Scene defined as an array of spheres (radius, position, emission, color, material) 
/*
  Sphere spheres[] = {
  Sphere(1e5, Vector( 1e5+1,40.8,81.6), Vector(0.0,0.0,0.0),Vector(.75,.25,.25),0),//Left red
  Sphere(1e5, Vector(-1e5+99,40.8,81.6),Vector(0.0,0.0,0.0),Vector(.25,.25,.75),0),//Right blue
  Sphere(1e5, Vector(50,40.8, 1e5),     Vector(0.0,0.0,0.0),Vector(.75,.75,.75),0),//Back 
  Sphere(1e5, Vector(50,40.8,-1e5+170), Vector(0.0,0.0,0.0),Vector(0.0,0.0,0.0),0),//Front 
  Sphere(1e5, Vector(50, 1e5, 81.6),    Vector(0.0,0.0,0.0),Vector(.75,.75,.75),0),//Botom white
  Sphere(1e5, Vector(50,-1e5+81.6,81.6),Vector(0.0,0.0,0.0),Vector(.75,.75,.75),0),//Top white
  Sphere(16.5,Vector(27,16.5,47),       Vector(0.0,0.0,0.0),Vector(1,1,1)*.999, 1),//Mirror 
  Sphere(16.5,Vector(73,16.5,78),       Vector(0.0,0.0,0.0),Vector(1,1,1)*.999, 2),//Glass 
  Sphere(600, Vector(50,681.6-.27,81.6),Vector(12,12,12),  Vector(0.0,0.0,0.0), 0) //Light 
  }; 
*/

Texture* damier1 = new TextureDamier(Vector(0.75,0.2,0.2),Vector(0.2,0.2,0.75));
Texture* fct1 = new TextureFct(Vector(1,0.2,0.2),Vector(0.2,0.8,1));
Texture* blanc = new TextureDefault(Vector(0.75,0.75,0.75));
Texture* rouge = new TextureDefault(Vector(0.75,0.25,0.25));
Texture* vert = new TextureDefault(Vector(0.25,0.75,0.25));
Texture* bleu = new TextureDefault(Vector(0.25,0.75,0.25));
Texture* noir = new TextureDefault(Vector(0,0,0));

Texture* transp = new TextureDefault(Vector(1,1,1)*.999);

Texture* tmp1 = new TextureDefault(Vector(0.2,0.1,0.75));
Texture* tmp2 = new TextureDefault(Vector(0.2,0.8,1));
Texture* tmp3 = new TextureDefault(Vector(0.3,0.8,0.1));
Texture* tmp4 = new TextureDefault(Vector(1,0.8,0.1));



/*************************************************************
 * Deux Tore orientés et de la réfraction
 ************************************************************/

Node* scene = new Or(
		     new Or(//scene
			    new Or(
				   new Or(//environnement
					  new Box(Parameters(Vector(0.0,0.0,0.0),damier1,0), Vector(-1,-1, -20), Vector(101,1,450)),//bas
					  new Box(Parameters(Vector(0.0,0.0,0.0),blanc,0), Vector(-1,80, -20), Vector(101,81,450))//haut
					  ),
				   new Or(
					  new Or(
						 new Box(Parameters(Vector(0.0,0.0,0.0),bleu,0), Vector(0,-1, -20), Vector(0,81,351)),//gauche
						 new Box(Parameters(Vector(0.0,0.0,0.0),rouge,0), Vector(100,-1, -20), Vector(100,81,351))//droite
						 ),
					  new Or(
						 new Box(Parameters(Vector(0.0,0.0,0.0),vert,1), Vector(-1,-1, 300), Vector(101,81,300)),//fond
						 new Box(Parameters(Vector(0.0,0.0,0.0),blanc,1), Vector(-1,-1, 0), Vector(101,81,0))//avant
						 )
					  )
				   ),//objets				   
			    new Or(
				   new Tore(Parameters(Vector(0.0,0.0,0.0),rouge,0), Vector(60,40, 80), Vector(1,1,1),20,5),
				   new Tore(Parameters(Vector(0.0,0.0,0.0),vert,0), Vector(51,50, 80), Vector(0,0,1),20,3)
				   )
		     
			    ),//lumiere
		     new Sphere(Parameters(Vector(12.0,12.0,12.0),noir,0), Vector(25, 95, 81.6), 20)
		     //new Sphere(Parameters(Vector(12.0,12.0,12.0),Vector(0,0,0),0), Vector(50, 78, 81.6), 4)
		     );


/*************************************************************
 * Texture bois et Texture damier
 ************************************************************/
/*
  Node* scene = new Or(
  new Or(//scene
  new Or(
  new Or(//environnement
  new Box(Parameters(Vector(0.0,0.0,0.0),transp,1), Vector(-1,-1, -20), Vector(101,1,450)),//bas
  new Box(Parameters(Vector(0.0,0.0,0.0),blanc,0), Vector(-1,80, -20), Vector(101,81,450))//haut
  ),
  new Or(
  new Or(
  new Box(Parameters(Vector(0.0,0.0,0.0),bleu,0), Vector(0,-1, -20), Vector(0,81,351)),//gauche
  new Box(Parameters(Vector(0.0,0.0,0.0),rouge,0), Vector(100,-1, -20), Vector(100,81,351))//droite
  ),
  new Or(
  new Box(Parameters(Vector(0.0,0.0,0.0),vert,0), Vector(-1,-1, 300), Vector(101,81,300)),//fond
  new Box(Parameters(Vector(0.0,0.0,0.0),blanc,0), Vector(-1,-1, 0), Vector(101,81,0))//avant
  )
  )
  ),//objets				   
  new Sub(
  new Sub(
  new Box(Parameters(Vector(0.0,0.0,0.0),damier1,0), Vector(20,30, 50), Vector(80,50,90)),
  new Box(Parameters(Vector(0.0,0.0,0.0),fct1,0), Vector(50,29, 61), Vector(75,61,91))
  ),
  new Sphere(Parameters(Vector(0.0,0.0,0.0),tmp3,0), Vector(50,50, 80), 15)
  )
  ),//lumiere
  new Sphere(Parameters(Vector(12.0,12.0,12.0),noir,0), Vector(25, 95, 81.6), 20)
  );
*/
/*************************************************************
 * Intersection, Union
 ************************************************************/
/*
  Node* scene = new Or(
  new Or(//scene
  new Or(
  new Or(//environnement
  new Box(Parameters(Vector(0.0,0.0,0.0),blanc,0), Vector(-1,-1, -20), Vector(101,1,450)),//bas
  new Box(Parameters(Vector(0.0,0.0,0.0),blanc,0), Vector(-1,90, -20), Vector(101,91,450))//haut
  ),
  new Or(
  new Or(
  new Box(Parameters(Vector(0.0,0.0,0.0),bleu,0), Vector(0,-1, -20), Vector(0,91,351)),//gauche
  new Box(Parameters(Vector(0.0,0.0,0.0),rouge,0), Vector(100,-1, -20), Vector(100,91,351))//droite
  ),
  new Or(
  new Box(Parameters(Vector(0.0,0.0,0.0),vert,0), Vector(-1,-1, 300), Vector(101,91,300)),//fond
  new Box(Parameters(Vector(0.0,0.0,0.0),blanc,0), Vector(-1,-1, 0), Vector(101,91,0))//avant
  )
  )
  ),//objets				   
  new And(
  new Or(
  new Sphere(Parameters(Vector(0.0,0.0,0.0),tmp2,0), Vector(40,40, 80), 30),
  new Sphere(Parameters(Vector(0.0,0.0,0.0),tmp1,0), Vector(61,40, 80), 27)
  ),
  new Sphere(Parameters(Vector(0.0,0.0,0.0),tmp3,0), Vector(50,41, 80), 30)
  )
			    
  ),//lumiere
  new Sphere(Parameters(Vector(12.0,12.0,12.0),noir,0), Vector(25, 95, 81.6), 20)

  );
*/

/*************************************************************
 * Cylindre, Cone, Tore
 ************************************************************/
/*
  Node* scene = new Or(
  new Or(//scene
  new Or(
  new Or(//environnement
  new Box(Parameters(Vector(0.0,0.0,0.0),blanc,0), Vector(-1,-1, -20), Vector(101,1,450)),//bas
  new Box(Parameters(Vector(0.0,0.0,0.0),blanc,0), Vector(-1,90, -20), Vector(101,91,450))//haut
  ),
  new Or(
  new Or(
  new Box(Parameters(Vector(0.0,0.0,0.0),bleu,0), Vector(0,-1, -20), Vector(0,91,351)),//gauche
  new Box(Parameters(Vector(0.0,0.0,0.0),rouge,0), Vector(100,-1, -20), Vector(100,91,351))//droite
  ),
  new Or(
  new Box(Parameters(Vector(0.0,0.0,0.0),vert,0), Vector(-1,-1, 300), Vector(101,91,300)),//fond
  new Box(Parameters(Vector(0.0,0.0,0.0),blanc,0), Vector(-1,-1, 0), Vector(101,91,0))//avant
  )
  )
  ),//objets				   
  new Or(
  new Or(
  new Cylinder(Parameters(Vector(0.0,0.0,0.0),tmp1,0), Vector(50,10, 100), Vector(20,30,100),15),
  new Cone(Parameters(Vector(0.0,0.0,0.0),tmp2,0), Vector(80,10, 80), Vector(50,40,70),15)
  ),
  new Tore(Parameters(Vector(0.0,0.0,0.0),tmp3,0), Vector(50,50, 40),Vector(1,1,1), 25,5)
  )
			    
  ),//lumiere
  new Sphere(Parameters(Vector(12.0,12.0,12.0),noir,0), Vector(25, 105, 81.6), 20)

  );

*/
/*************************************************************
 * Opérateurs
 ************************************************************/

/*
  Node* scene = new Or(
  new Or(//scene
  new Or(
  new Or(//environnement
  new Box(Parameters(Vector(0.0,0.0,0.0),blanc,0), Vector(-1,-1, -20), Vector(101,1,450)),//bas
  new Box(Parameters(Vector(0.0,0.0,0.0),blanc,0), Vector(-1,90, -20), Vector(101,91,450))//haut
  ),
  new Or(
  new Or(
  new Box(Parameters(Vector(0.0,0.0,0.0),bleu,0), Vector(0,-1, -20), Vector(0,91,351)),//gauche
  new Box(Parameters(Vector(0.0,0.0,0.0),rouge,0), Vector(100,-1, -20), Vector(100,91,351))//droite
  ),
  new Or(
  new Box(Parameters(Vector(0.0,0.0,0.0),vert,0), Vector(-1,-1, 300), Vector(101,91,300)),//fond
  new Box(Parameters(Vector(0.0,0.0,0.0),blanc,0), Vector(-1,-1, 0), Vector(101,91,0))//avant
  )
  )
  ),//objets				   
  new Or(
  new Or(
  new Sphere(Parameters(Vector(0.0,0.0,0.0),tmp1,0), Vector(80,49, 80), 20),
  new Cylinder(Parameters(Vector(0.0,0.0,0.0),tmp3,0), Vector(80,0,80),Vector(80,30,80),5)
  ),
  new Or(
  new Sub(new Sphere(Parameters(Vector(0.0,0.0,0.0),tmp2,0), Vector(50,50,40),25),
  new Tore(Parameters(Vector(0.0,0.0,0.0),tmp4,0), Vector(50,50, 40),Vector(1,1,1), 25,5)
  ),
					  
  new And(
  new Box(Parameters(Vector(0.0,0.0,0.0),tmp4,0), Vector(20,10, 80),Vector(50,40,110)),
  new Sphere(Parameters(Vector(0.0,0.0,0.0),tmp4,0), Vector(20,10, 80),30)
  )
  )
  )
  ),//lumiere
  new Sphere(Parameters(Vector(12.0,12.0,12.0),noir,0), Vector(25, 105, 81.6), 20)

  );


*/

/*************************************************************
 * Une loupe => mauvais rendu (mauvaise gestion de la transparence ?
 ************************************************************/
/*
  Node* scene = new Or(
  new Or(//scene
  new Or(
  new Or(//environnement
  new Box(Parameters(Vector(0.0,0.0,0.0),damier1,0), Vector(-1,-1, -20), Vector(101,1,450)),//bas
  new Box(Parameters(Vector(0.0,0.0,0.0),blanc,0), Vector(-1,80, -20), Vector(101,81,450))//haut
  ),
  new Or(
  new Or(
  new Box(Parameters(Vector(0.0,0.0,0.0),bleu,0), Vector(0,-1, -20), Vector(0,81,351)),//gauche
  new Box(Parameters(Vector(0.0,0.0,0.0),rouge,0), Vector(100,-1, -20), Vector(100,81,351))//droite
  ),
  new Or(
  new Box(Parameters(Vector(0.0,0.0,0.0),transp,1), Vector(-1,-1, 300), Vector(101,81,300)),//fond
  new Box(Parameters(Vector(0.0,0.0,0.0),blanc,0), Vector(-1,-1, 0), Vector(101,81,0))//avant
  )
  )
  ),//objets				   
  new Or(
  // new Cone(Parameters(Vector(0.0,0.0,0.0),transp,2), Vector(50,1, 80),Vector(50,50,80),40),
  // new Sphere(Parameters(Vector(0.0,0.0,0.0),vert,0), Vector(50,30, 10),10)
  new Or(
  new And(
  new Sphere(Parameters(Vector(0.0,0.0,0.0),transp,2), Vector(50,33, 80-96),100),
  new Sphere(Parameters(Vector(0.0,0.0,0.0),transp,2), Vector(50,33, 80+96),100)
  ),
  //new Sphere(Parameters(Vector(0.0,0.0,0.0),transp,0), Vector(50,50, 80),10),

  new Tore(Parameters(Vector(0.0,0.0,0.0),rouge,0), Vector(50,33, 80), Vector(0,0,1),26,2)
  ),
  new Box(Parameters(Vector(0.0,0.0,0.0),tmp3,0), Vector(50,25, 30),Vector(90,60,40))
  )
		     
  ),//lumiere
  new Sphere(Parameters(Vector(12.0,12.0,12.0),noir,0), Vector(25, 95, 81.6), 20)
  //new Sphere(Parameters(Vector(12.0,12.0,12.0),noir,0), Vector(50,701.6-.27,81.6), 600)
		     
  );
*/
/*************************************************************
 * Un verre de lait => ne donne pas grand chose au rendu
 ************************************************************/
/*
  Node* scene = new Or(
  new Or(//scene
  new Or(
  new Or(//environnement
  new Box(Parameters(Vector(0.0,0.0,0.0),damier1,0), Vector(-1,-1, -20), Vector(101,1,450)),//bas
  new Box(Parameters(Vector(0.0,0.0,0.0),blanc,0), Vector(-1,80, -20), Vector(101,81,450))//haut
  ),
  new Or(
  new Or(
  new Box(Parameters(Vector(0.0,0.0,0.0),bleu,0), Vector(0,-1, -20), Vector(0,81,351)),//gauche
  new Box(Parameters(Vector(0.0,0.0,0.0),rouge,0), Vector(100,-1, -20), Vector(100,81,351))//droite
  ),
  new Or(
  new Box(Parameters(Vector(0.0,0.0,0.0),vert,0), Vector(-1,-1, 300), Vector(101,81,300)),//fond
  new Box(Parameters(Vector(0.0,0.0,0.0),blanc,0), Vector(-1,-1, 0), Vector(101,81,0))//avant
  )
  )
  ),//objets
  new Or(
  new Sub(
				    
  new Cylinder(Parameters(Vector(0.0,0.0,0.0),transp,2), Vector(50,2, 40), Vector(50,40,40),20),
  new Cylinder(Parameters(Vector(0.0,0.0,0.0),transp,2), Vector(50,6, 40), Vector(50,41,40),19)
  ),
  new Cylinder(Parameters(Vector(0.0,0.0,0.0),blanc,0), Vector(50,6.001, 120), Vector(50,20,120),18.999)
  )
		     
  ),//lumiere
  new Sphere(Parameters(Vector(12.0,12.0,12.0),noir,0), Vector(25, 95, 81.6), 20)
  );*/


/*!
  \brief Clamp value to unity.
*/
inline double clamp(double x){ return x<0 ? 0 : x>1 ? 1 : x; } 

/*!
  \brief Convert to int and apply a gamma power of 2.2.
*/
inline int toInt(double x){ return int(pow(clamp(x),1/2.2)*255+.5); } 

/*!
  \brief Compute the intersection between the ray and the scene.
*/
//#define M_PI 3.141592654 //déjà dans cmath

inline void CoordinateSystem(const Vector& v1, Vector& v2, Vector& v3)
{
  if (fabs(v1[0]) > fabs(v1[1]))
    {
      double invLen = 1.0 / (v1[0] * v1[0] + v1[2] * v1[2]);
      v2 = Vector(-v1[2] * invLen, 0.0, v1[0] * invLen);
    }
  else
    {
      double invLen = 1.0 / sqrt(v1[1] * v1[1] + v1[2] * v1[2]);
      v2 = Vector(0.0, v1[2] * invLen, -v1[1] * invLen);
    }

  v3 = v1 / v2;
}

/*!
  \brief Compute the radiance of a given ray.
  \param r The ray.
  \brief depth Recursion depth.
*/
Vector radiance(const Ray &r, int depth)
{ 

  Intersections l;                               // distance to IntersectSceneion 
  int id=0;                               // id of IntersectSceneed object 
  // if (!IntersectScene(r, t, id)) return Vector(0.0,0.0,0.0); 
  if(!scene->Intersect(r,l))  return Vector(0.0,0.0,0.0); 

  Parameters obj(l.first());
  double t =    obj.getDistance() ;


  Vector x=r(t);
  //Vector n=Normalized(x-obj.getPosition());
  Vector n=obj.getNormal();
  Vector nl=n*(r.Direction())<0?n:n*-1.0;

  Vector f=obj.getColor();
  double p=obj.getMaxColorChannel();

  if (++depth>D) //if (Random()<p) f=f*(1/p); else
    return obj.getEmission(); //R.R. 
  if (obj.getReflection() == 0) // Diffuse
    {    

      double r1=2*M_PI*Random(), r2=Random(), r2s=sqrt(r2); 
      Vector w=nl;
      Vector u;
      Vector v;

      CoordinateSystem(w,u,v);

      Vector d = Normalized(u*cos(r1)*r2s + v*sin(r1)*r2s + w*sqrt(1-r2)); 
      Vector result = obj.getEmission() + f.Scale(radiance(Ray(x,d),depth));
      return result; 
    } 
  else if (obj.getReflection() == 1)  // Specular
    {
      return obj.getEmission() + f.Scale(radiance(Ray(x,r.Direction()-n*2*(n*r.Direction())),depth)); 
    }
  else // Refractive
    {

      Ray reflRay(x, r.Direction()-n*2*n*(r.Direction()));    
      bool into = n*nl>0;                // Ray from outside going in? 
      double nc=1, nt=1.5, nnt=into?nc/nt:nt/nc, ddn=r.Direction()*nl, cos2t; 
      if ((cos2t=1-nnt*nnt*(1-ddn*ddn))<0)    // Total internal reflection 
	return obj.getEmission() + f.Scale(radiance(reflRay,depth)); 
      Vector tdir = Normalized(r.Direction()*nnt - n*((into?1:-1)*(ddn*nnt+sqrt(cos2t)))); 
      double a=nt-nc, b=nt+nc, R0=a*a/(b*b), c = 1-(into?-ddn:tdir*n); 
      double Re=R0+(1-R0)*c*c*c*c*c,Tr=1-Re,P=.25+.5*Re,RP=Re/P,TP=Tr/(1-P); 

      return obj.getEmission() + f.Scale(depth>2 ? (Random()<P ?   // Russian roulette 
						    radiance(reflRay,depth)*RP:radiance(Ray(x,tdir),depth)*TP) : 
					 radiance(reflRay,depth)*Re+radiance(Ray(x,tdir),depth)*Tr); 
    }

} 


#define IDENT(n) for (int i = 0; i < n; ++i) printf("    ")
Node* sc;
Node* crt;
void printNode(json_value *value, int ident = 0)
{
 
  bool newNode = false;
  IDENT(ident);
  if (value->name){
    printf("\"%s\" = ", value->name);
    if(value->name=="node"){
      newNode = true;
    }
  }
  switch(value->type)
    {
    case JSON_NULL:
      printf("null\n");
      break;
    case JSON_OBJECT:
    case JSON_ARRAY:
      printf(value->type == JSON_OBJECT ? "{\n" : "[\n");
      for (json_value *it = value->first_child; it; it = it->next_sibling)
	{
	  printNode(it, ident + 1);
	}
      IDENT(ident);
      printf(value->type == JSON_OBJECT ? "}\n" : "]\n");
      break;
    case JSON_STRING:
      printf("\"%s\"\n", value->string_value);
      if(newNode){
	if(value->string_value=="Or"){ sc = new Or();}
	//	else if(value->string_value=="Sphere"){  sc = new Sphere();}
      }
      break;
    case JSON_INT:
      printf("%d\n", value->int_value);
      break;
    case JSON_FLOAT:
      printf("%f\n", value->float_value);
      break;
    case JSON_BOOL:
      printf(value->int_value ? "true\n" : "false\n");
      break;
    }

}

bool parse(char *source)
{

  char *errorPos = 0;
  char *errorDesc = 0;
  int errorLine = 0;
  block_allocator allocator(1 << 10);
  json_value *root = json_parse(source, &errorPos, &errorDesc, &errorLine, &allocator);
  if (root)
    {
      printNode(root);
      return true;
    }

  printf("Error at line %d: %s\n%s\n\n", errorLine, errorDesc, errorPos);
  return false;
}

static void * threadInit (void * p_data)
{
  return NULL;
}
static void * fn_render (void * ptr)
{
  thdata *data;            
  data = (thdata *) ptr;
  int nb = data->thread_no +1;

  int w=X; 
  int h=Y; 
  int tmp = h/NB_THREAD;

  int samps = S;
  Ray camera(Vector(50,52,295.6), Normalized(Vector(0,-0.042612,-1))); // camera pos, dir 
  //Ray camera(Vector(-120,-120,180), Normalized(Vector(1,1,-1.5))); // camera pos, dir 
  //Ray camera(Vector(-180,10,10), Normalized(Vector(1,-.1,-.1))); // camera pos, dir 
  Vector cx=Vector(w*.5135/h,0,0), cy=Normalized(cx/camera.Direction())*.5135;
  // printf("thread %d de %d à %d\n",nb,(tmp)*nb,(tmp)*(nb-1));

  for (int y=tmp * (nb-1); y<tmp * nb; y++)
    {  
      fprintf(stderr,"\rRendering (%d spp) %5.2f%%",samps*4,100.*y/(h-1)); 
      for (unsigned short x=0; x<w; x++)
	{
	  int i=(h-y-1)*w+x;
	  render.c[i]=Vector(0.0,0.0,0.0);
	  
	  for (int sy=0; sy<2; sy++)
	    {
	      // 2x2 subpixel rows 
	      for (int sx=0; sx<2; sx++)
		{
		  Vector r(0.0,0.0,0.0);
		  for (int s=0; s<samps; s++)
		    { 
		      double r1=2*Random(), dx=r1<1 ? sqrt(r1)-1: 1-sqrt(2-r1); 
		      double r2=2*Random(), dy=r2<1 ? sqrt(r2)-1: 1-sqrt(2-r2); 
		      Vector d = cx*( ( (sx+0.5 + dx)/2 + x)/w - 0.5) + cy*( ( (sy+0.5 + dy)/2 + y)/h - 0.5) + camera.Direction(); 
		      r += radiance(Ray(camera.Origin()+d*140,Normalized(d)),0)*(1./samps); 
		    } 
		  
		  render.c[i] += Vector(clamp(r[0]),clamp(r[1]),clamp(r[2]))*.25; 
		} 
	    } 
	  //	  std::cout<<"["<<x<<";"<<y<<"] "<<nb<<"=>"<<i<<" "<<render.c[i]<<std::endl;
	}
    }

  
  
  return NULL;
}

int main(int argc, char *argv[])
{ 

 
    

  std::string filename  = argc <= 2 ? "image.ppm" :argv[1];
  //    FILE *fp = fopen(argv[1], "rb");
  //if (fp) {


  // réccupération du contenu du fichier JSON

  /*    fseek(fp, 0, SEEK_END);
	int size = ftell(fp);
	fseek(fp, 0, SEEK_SET);
	std::vector<char> buffer(size + 1);
	fread (&buffer[0], 1, size, fp);
	fclose(fp);*/

  //if(parse(&buffer[0])){

  int i = 0;
  int ret = pthread_create (
			    & render.thread_render, NULL,threadInit, NULL
			    );

  if(!ret){
    time_t start = time(NULL);


    int w=X; 
    int h=Y;   
    render.c=new Vector[X*Y]; 

    thdata data[NB_THREAD]; 
    for (i = 0; i < NB_THREAD; i++){
      
      data[i].thread_no = i;
      ret = pthread_create (&render.thread_renders [i], NULL,fn_render, (void *) &data[i]);
    }

    for (i = 0; i < NB_THREAD; i++) {
      pthread_join (render.thread_renders [i], NULL);
    }
    pthread_join (render.thread_render, NULL);
    // Save image as PPM file
    //FILE *f = fopen("image.ppm", "w"); 
    FILE *f = fopen(filename.c_str(), "w"); 
    fprintf(f, "P3\n%d %d\n%d\n", w, h, 255); 
    for (int i=0; i<w*h; i++) 
      {
	fprintf(f,"%d %d %d ", toInt(render.c[i][0]), toInt(render.c[i][1]), toInt(render.c[i][2]));
      }

    fclose(f);
    delete []render.c;
    time_t end = time(NULL);
    double duree=difftime(end, start);
 
    printf("\nTemps de rendu : %f\n",duree);
  }
     

  return 0;
}
